package Injection is
   
   type Length is range 0 .. 1_000; 
   subtype Index is Length;
   type Permutation is array (Index range <>) of Index;
   
   procedure Invert
     (N   : in  Length;
      A   : in  Permutation;
      B   : out Permutation);
   --# derives B from A, N;
   --# pre  A'First = 0 and A'Last = N-1 and
   --#      B'First = 0 and B'Last = N-1 and
   --#      (for all J in Index range 0 .. N-1 => (0 <= A(J) and A(J) <= N-1))
   --#
   --#  -- A is injective:
   --#      and (for all J in Index range 0 .. N-1 =>
   --#            (for all K in Index range J+1 .. N-1 =>
   --#              (A(J) /= A(K))))
   --#
   --#  -- A is surjective:
   --#      and (for all M in Index range 0 .. N-1 =>
   --#            (for some K in Index range 0 .. N-1 =>
   --#              (A(K) = M)));
   --# post (for all J in Index range 0 .. N-1 => (B(A(J)) = J)) and
   --#      (for all J in Index range 0 .. N-1 => (0 <= B(J) and B(J) <= N-1))
   --#
   --#  -- B is injective:
   --#      and (for all J in Index range 0 .. N-1 =>
   --#            (for all K in Index range J+1 .. N-1 =>
   --#              (B(J) /= B(K))));

end Injection;
		      
   
