package body Amortized_Queue is

   function Copy (L : in List) return List
   --# pre Inv_List(L);
   --# return R => (Inv_List(R) and R.Size = L.Size);
   is
      --# hide Copy;
      Raw  : Data;
      Lres : List;
   begin
      for N in Index loop
         --# accept F, 23, Raw, "Array is completely initialized in the loop";
         Raw (N) := L.Raw (N);
         --# assert (for all K in Index range 0 .. N => (Raw (N) = L.Raw (N)));
      end loop;

      Lres.Free := L.Free;
      Lres.Head := L.Head;
      Lres.Raw  := Raw;
      Lres.Size := L.Size;
      --# accept F, 602, Raw, "Array is completely initialized in the loop";
      return Lres;
   end Copy;

   function Pop (L : in List) return List
   --# pre L.Size > 0 and Inv_List(L);
   --# return R => (Inv_List(R) and R.Size=L.Size-1);
   is
      --# hide Pop;
      Lres : List;
   begin
      Lres := Copy (L);

      if L.Free = No_Index then
         Lres.Raw (L.Head).Next := L.Head;
      else
         Lres.Raw (L.Head).Next := L.Free;
      end if;

      Lres.Free := L.Head;
      Lres.Head := L.Raw (L.Head).Next;
      Lres.Size := L.Size - 1;
      return Lres;
   end Pop;

   function Peek (L : in List) return Val
   --# pre L.Size > 0 and Inv_List(L);
   is
      --# hide Peek;
   begin
      return L.Raw (L.Head).Value;
   end Peek;

   procedure Add (L : in out List; V : in Val)
   --# derives L from L, V;
   --# pre  L.Size < Length'Last and Inv_List(L);
   --# post Inv_List(L) and L.Size = L~.Size + 1;
   is
      --# hide Add;
      Head : Index;
      Free : Extended_Index;
   begin
      Head := L.Free;
      Free := L.Raw (Head).Next;

      if Free = No_Index and then Head < Index'Last then
         Free := Head + 1;
      end if;

      if Free = Head then
         Free := No_Index;
      end if;

      L.Raw (Head) := Node'(Value => V, Next => L.Head);
      L.Head := Head;
      L.Free := Free;
      L.Size := L.Size + 1;
   end Add;

   function Push (L : in List; V : in Val) return List
   --# pre L.Size < Length'Last and Inv_List(L);
   --# return R => Inv_List(R) and R.Size = (L.Size) + 1;
   is
      Lres : List;
   begin
      Lres := Copy (L);
      Add(Lres, V);
      return Lres;
   end Push;

   procedure RevAppend (Lfront : in out List; Lrear : in out List)
   --# derives Lfront from Lfront, Lrear &
   --#         Lrear  from Lrear;
   --# pre Lfront.Size + Lrear.Size <= Length'Last and
   --#     Inv_List (Lrear) and
   --#     Inv_List(Lfront);
   --# post Inv_List(Lfront) and
   --#      Inv_List (Lrear) and
   --#      Lrear.Size = 0 and
   --#      Lfront.Size = Lfront~.Size + Lrear~.Size;
   is
      --# hide RevAppend;
      N      : Extended_Index;
      IArray : IntArray;
      I      : Extended_Index;
      Lres   : List;
   begin
      Lres := List'(Head => 0,
                    Raw  => No_Data,
                    Free => 1,
                    Size => 0);
      N := Lrear.Head;

      while N /= No_Index loop
         --# assert True;
         Add (Lres, Lrear.Raw (N).Value);
         N := Lrear.Raw (N).Next;
      end loop;

      N := Lfront.Head;
      I := Extended_Index'First;

      while N /= No_Index loop
         --# assert True;
         I := I + 1;
         --# accept F, 23, IArray, "Array is initialized in the loop";
         IArray (I) := Lfront.Raw (N).Value;
         N := Lfront.Raw (N).Next;
      end loop;

      while I >= Index'First loop
         --# assert True;
         --# accept F, 504, IArray, "Array is initialized in the loop";
         Add (Lres, IArray (I));
         I := I - 1;
      end loop;

      LRear.Head := 0;
      LRear.Free := 1;
      LRear.Size := 0;
      LFront     := Lres;
      LRear.Raw  := No_Data;
      --# accept F, 602, Lfront, IArray, "Array is initialized in the loop";
   end RevAppend;

   function Tail (Q : Queue) return Queue is
      My_Front : List;
      My_Rear  : List;
   begin
      My_Front := Pop (Q.Front);
      My_Rear  := Copy (Q.Rear);

      if My_Front.Size < My_Rear.Size then
         RevAppend (My_Front, My_Rear);
      end if;

      return Queue'(Front => My_Front, Rear => My_Rear);
   end Tail;

   function Enqueue (Q : in Queue; V : in Val) return Queue is
      My_Front : List;
      My_Rear  : List;
   begin
      My_Front := Copy (Q.Front);
      My_Rear  := Push (Q.Rear, V);

      if My_Front.Size < My_Rear.Size then
         RevAppend (My_Front, My_Rear);
      end if;

      return Queue'(Front => My_Front, Rear => My_Rear);
   end Enqueue;

   function Front (Q : Queue) return Val is
   begin
      return Peek (Q.Front);
   end Front;

end Amortized_Queue;
