package Amortized_Queue is

   type Length is range 0 .. 1_000;
   subtype Extended_Index is Length;
   subtype Index is Extended_Index range 1 .. 1_000;

   No_Index : constant := 0;

   subtype Val is Integer range -2 ** 31 .. 2 ** 31 - 1;
   type IntArray is array(Index) of Val;

   type Node is record
      Value : Val;
      Next  : Extended_Index;
   end record;

   type Data is array (Index) of Node;

   No_Data : constant Data := Data'(others => Node'(Value => 0, Next => 0));

   type List is record
      Head : Extended_Index;
      Free : Extended_Index;
      Raw  : Data;
      Size : Length;
   end record;

   --# function Inv_List (L : in List) return Boolean;
   --  Invariant of list is that all array indexes in L.Raw are reachable from
   --  either Head (in list) or Free (available):
   --    - reachability from Head is through Next fields, until the value
   --      No_Index.
   --    - reachability from Free is through Next fields, until the value loops
   --      on itself with field Next, or it is No_Index in which case all
   --      following indexes of the array are free.

   type Queue is record
      Front : List;
      Rear  : List;
   end record;

   function Front (Q : in Queue) return Val;
   --# pre Q.Front.Size + Q.Rear.Size > 0 and
   --#     Q.Front.Size + Q.Rear.Size <= Length'Last and
   --#     Q.Front.Size >= Q.Rear.Size and
   --#     Inv_List (Q.Front) and
   --#     Inv_List (Q.Rear);

   function Tail (Q : in Queue) return Queue;
   --# pre Q.Front.Size + Q.Rear.Size > 0 and
   --#     Q.Front.Size + Q.Rear.Size <= Length'Last and
   --#     Q.Front.Size >= Q.Rear.Size and
   --#     Inv_List (Q.Front) and
   --#     Inv_List (Q.Rear);
   --# return R => Q.Front.Size + Q.Rear.Size <= Length'Last and
   --#             Q.Front.Size >= Q.Rear.Size and
   --#             Inv_List (Q.Front) and
   --#             Inv_List (Q.Rear);

   function Enqueue (Q : in Queue; V : in Val) return Queue;
   --# pre Q.Front.Size + Q.Rear.Size < Length'Last and
   --#     Q.Front.Size + Q.Rear.Size <= Length'Last and
   --#     Q.Front.Size >= Q.Rear.Size and
   --#     Inv_List (Q.Front) and
   --#     Inv_List (Q.Rear);
   --# return R => Q.Front.Size + Q.Rear.Size <= Length'Last and
   --#             Q.Front.Size >= Q.Rear.Size and
   --#             Inv_List (Q.Front) and
   --#             Inv_List (Q.Rear);

end Amortized_Queue;
